#include "acctd-config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <syslog.h>
#include <sysexits.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
extern int flock(int, int);
#include <limits.h>
#include <time.h>
#include <fcntl.h>

#ifdef HAVE_DB_185_H
#include <db_185.h>
#else
#include <db.h>
#endif

#include "own_wtmp.h"

#ifdef __STDC__
#include <stdarg.h>	/* ANSI C, variable length args */
#else
#include <varargs.h>	/* has 'vararg' definitions */
#endif

char *pname;
char *utmp_path = TAC_PLUS_UTMPFILE;
char *dir_path = NULL;
int utmp_file = -1;
time_t scan_time = SCANTIME;
time_t reinit = REINIT;
char *killer = ACTPROG;
char *dbfile = TAC_PLUS_DBFILE;
#ifdef ACCTD_NOTIFY
char *notifier = NOTIFYPROG;
int  notifytime = NOTIFYTIME;
#endif

static int debug = 0;
int nodetach = 0;

DB *db = NULL;
DBT key[1], val[1];

time_t told, tnow;
#define pidfilename RUNPATH "/acctd.pid"

#ifdef __STDC__
static void report(int priority, char *fmt, ...)
#else /* VARARGS2 */
static void report(priority, fmt, va_alist)
int priority;
char *fmt;
va_dcl                        /* no terminating semi-colon */
#endif
{
  char msg[1024];             /* temporary string */
  va_list ap;

#ifdef __STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif
  vsnprintf(msg, sizeof(msg), fmt, ap);
  va_end(ap);

  syslog(priority, "%s", msg);
  if (debug) printf("%s\n", msg);
}

#if defined(STRICT_TIME_CHECK) || defined(PARROT)
#include "tac_plus.h"
extern double parrot_check(char *);
#endif

#ifdef PARROT
static double parrot_off=PARROT_OFF; /* avoid rounding errors */
#endif

#ifdef STRICT_TIME_CHECK
extern int cfg_read_config(char *);
extern int set_time_status(char *, struct authen_data *, struct passwd *, int);
int initialised = 0;
int parse_only = 0;
unsigned long bindaddress;
char *cfgfile=TAC_PLUS_CFGFILE;

int strict_time_check(char *name, char *nas, char *port) {
  struct authen_data a1;
  struct identity a2;
  struct passwd pw;

  bzero(&a2, sizeof(a2));
  a2.username = name;
  a2.NAS_name = nas;
  a2.NAS_port = port;

  bzero(&a1, sizeof(a1));
  a1.NAS_id = &a2;

  bzero(&pw, sizeof(pw));
  pw.pw_name = name;
  pw.pw_uid = NOPUID;
  pw.pw_gid = NOPGID;
  pw.pw_dir = NOPHOME;
  pw.pw_shell = NOPSHELL;

  return set_time_status(name, &a1, &pw, TAC_PLUS_RECURSE);
}
#endif

void usage()
{
  fprintf(stderr, "\
Usage: %s [-I] [-d] [-w file] [-r [dir]] [-t time] [-a [time]] [-k path]\n\
 -I          - don't detach from controlling terminal\n\
 -d          - run in debug mode (do not daemonize)\n\
 -w <file>   - use utmp file <file> instead of %s\n\
 -u <file>   - use user database file <file> instead of %s\n\
 -r [<dir>]  - place restrict files in <dir> (%s as default) directory\n\
 -t <time>   - scan wtmp each <time> seconds instead of %d\n\
 -a <time>   - kill users with great than <time> seconds on-line time\n\
               (if <time>==0, then never) instead of %d\n\
 -k <path>   - path to killer program instead of %s\n"
#ifdef ACCTD_NOTIFY
" -n <path>   - path to notifier program instead of %s\n\
 -o <time>   - send notification when left <time> minutes instead of %d\n"
#endif
, pname, TAC_PLUS_UTMPFILE, TAC_PLUS_DBFILE, TAC_PLUS_DENYDIR, SCANTIME, REINIT, ACTPROG
#ifdef ACCTD_NOTIFY
, NOTIFYPROG, NOTIFYTIME
#endif
);
  exit(EX_USAGE);
}

void hangup(int s) {
  report(LOG_INFO, "got %d signal -- exiting", s);
  if (db)
    (*db->close)(db);
  closelog();
  unlink(pidfilename);
  exit(EX_OK);
}

void killuser(UTMP_T *w, int act) {
  int pid;

  if ((pid = fork())) {
    int c;
    waitpid(pid, &c, 0);
    return;
  }
#ifdef HAVE_DAEMON
  daemon(1, 0);
#else
  setsid();
#endif
  if ((pid = fork())) {
    sleep(1);
    exit(EX_OK);
  }
#ifdef HAVE_DAEMON
  daemon(1, 0);
#else
  setsid();
#endif
  if (dir_path) {
    time_t t;
    char lck[256];
    FILE *f;

    time(&t);
    if (dir_path && act) {
      if(sizeof(lck) <= strlen(dir_path) + strlen(w->ut_user) + 1)
	report(LOG_ERR, "can't create deny file for %s: internal buffer overflow", w->ut_user);
      else {
	sprintf(lck, "%s/%s", dir_path, w->ut_user);
	if ((f = fopen(lck, "w"))) {
	  fprintf(f, "\n*** Sorry, your time limit is exceeded at %s", ctime(&t));
	  fclose(f);
	} else
	  report(LOG_ERR, "can't create deny file %s: %s", lck, strerror(errno));
      }
    }
  }
  execl(killer, "-killuser", w->ut_user, w->ut_line, w->ut_host, NULL);
  report(LOG_ERR, "can't exec %s %s %s %s: %s", killer, w->ut_user, w->ut_line, w->ut_host, strerror(errno));
  exit(EX_OSERR);
}

#ifdef ACCTD_NOTIFY
void notifyuser(UTMP_T *w) {
  int pid;

  if ((pid = fork())) {
    int c;
    waitpid(pid, &c, 0);
    return;
  }
#ifdef HAVE_DAEMON
  daemon(1, 0);
#else
  setsid();
#endif
  if ((pid = fork())) {
    sleep(1);
    exit(EX_OK);
  }
#ifdef HAVE_DAEMON
  daemon(1, 0);
#else
  setsid();
#endif
  execl(notifier, "-notifier", w->ut_user, w->ut_line, w->ut_host, inet_ntoa(w->ut_addr), NULL);
  report(LOG_ERR, "can't exec %s %s %s %s %s: %s", notifier, w->ut_user, w->ut_line, w->ut_host, inet_ntoa(w->ut_addr), strerror(errno));
  exit(EX_OSERR);
}
#endif

int db_update(DB *dbp, time_t t, UTMP_T *w) {
#ifdef DBF_1
  TQ_st tq;
  TQ_st *tqp;
#else
  time_t tq;
#endif

  key->data = w->ut_user;
  key->size = strlen(w->ut_user) + 1;
  if ((*db->get) (dbp, key, val, NULL))
    return -1;

#ifdef DBF_1
  tqp = (TQ_st *) val->data;
  memcpy(&tq, tqp, sizeof(tq));
  if (tq.tl != TIME_LIMIT_OFF)
    tq.tll += t;
  if (tq.dl != TIME_LIMIT_OFF)
    tq.dll += t;
  tq.tot = tqp->tot - t;
#ifdef PARROT
  if (tq.parrot != parrot_off) {
    double veloc = parrot_check(w->ut_user);
    tq.parrot += t * veloc / PARROT_INTERVAL;
    if (tq.parrot + tq.parrot_credit <= 0.0) {
      killuser(w, 1);
      report(LOG_INFO, "user %s on line %s nas %s parrot limit reached", w->ut_user, w->ut_line, w->ut_host);
      goto e1;
    }
#ifdef ACCTD_NOTIFY
    if (tq.parrot + tq.parrot_credit <= notifytime * veloc
	&& tq.parrot + tq.parrot_credit > (notifytime-1) * veloc) {
      /* send notify to user */
      report(LOG_INFO, "notifying user %s on line %s nas %s about expiration of parrot limit", w->ut_user, w->ut_line, w->ut_host);
      notifyuser(w);
    }
#endif
  }
#endif
  if (tq.dll <= 0) {
    if ((tq.dl != TIME_LIMIT_NO) && (tq.dl != TIME_LIMIT_OFF)) { /* was (tq.dl >= 0) */
      killuser(w, 1);
      report(LOG_INFO, "user %s on line %s nas %s day time limit reached", w->ut_user, w->ut_line, w->ut_host);
      goto e1;
    }
  }
  if (tq.tll <= 0) {
    if ((tq.tl != TIME_LIMIT_NO) && (tq.tl != TIME_LIMIT_OFF)) { /* was (tq.tl >= 0) */
      killuser(w, 1);
      report(LOG_INFO, "user %s on line %s nas %s total time limit reached", w->ut_user, w->ut_line, w->ut_host);
    }
  }
e1:
#else
  tq = *(time_t *)val->data;
  tq += t;
  if (tq < 0) {
    killuser(w, 1);
    report(LOG_INFO, "user %s on line %s nas %s time limit reached", w->ut_user, w->ut_line, w->ut_host);
    tq = 0;
  }
#endif
  val->data = (void *) &tq;
  val->size = sizeof(tq);
  return (*db->put) (db, key, val, NULL);
}


int main(int ac, char *av[]) {
  int ch;
  UTMP_T w;
#ifndef HAVE_DAEMON
  int nf;
#endif
  FILE *pidfile;
  pid_t pid,pid_old; /* Our pid */

  pname = strrchr(av[0], '/');
  pname = pname ? ++pname : av[0];

  opterr = 0;
  while ((ch = getopt(ac, av, "Iu:dw:r:t:a:k:"
#ifdef ACCTD_NOTIFY
		      "n:o:"
#endif
		      )) != EOF) {
    if(ch == ':')
      ch = optopt;
    switch (ch) {
    case 'I':
      nodetach = 1;
      break;
    case 'w':
      if(!optarg)
	usage();
      utmp_path = optarg;
      break;
    case 'u':
      if(!optarg)
	usage();
      dbfile = optarg;
      break;
    case 'r':
      dir_path = optarg ? optarg : TAC_PLUS_DENYDIR;
      break;
    case 't':
      if(!optarg)
	usage();
      scan_time = atoi(optarg);
      break;
    case 'a':
      if(!optarg)
	usage();
      reinit = atoi(optarg);
      break;
    case 'k':
      if(!optarg)
	usage();
      killer = optarg;
      break;
#ifdef ACCTD_NOTIFY
    case 'n':
      if(!optarg)
	usage();
      notifier = optarg;
      break;
    case 'o':
      if(!optarg)
	usage();
      notifytime = atoi(optarg);
      break;
#endif
    case 'd':
      debug = 1;
      break;
    default:
      usage();
      break;
    }
  }
  ac -= optind;
  av += optind;

  openlog(pname, LOG_PID | LOG_NDELAY, LOG_FACIL);

  utmp_file = open(utmp_path, O_RDONLY);
  if (utmp_file < 0) {
    report(LOG_ERR, "unable to open wtmp file %s: %s", utmp_path, strerror(errno));
    exit(EX_OSFILE);
  }
  signal(SIGHUP, hangup);
  signal(SIGINT, hangup);
  signal(SIGQUIT, hangup);
  signal(SIGTERM, hangup);

  report(LOG_INFO, "started (version %s)", ACCTD_VERSION);

  if (!debug && !nodetach) {
#ifdef HAVE_DAEMON
    daemon(1, 0);
#else
    nf = open("/dev/null", O_RDWR, 0600);
    dup2(nf, 0);
    dup2(nf, 1);
    dup2(nf, 2);
    close(nf);
    setsid();
    if (fork())
      exit(0);
#endif
  }

#if defined(STRICT_TIME_CHECK) || defined(PARROT)
  parser_init();

  if (initialised)
    cfg_clean_config();

  report(LOG_INFO, "Reading config");

  if (!cfgfile) {
    report(LOG_ERR, "no config file specified");
    return 1;
  }

  if (cfg_read_config(cfgfile)) {
    report(LOG_ERR, "Parsing %s", cfgfile);
    return 1;
  }
#endif

/* write pid to file -- patch from Ivan V. Sidorenko */
  pid = getpid();
  if ((pidfile = fopen(pidfilename, "r")) != NULL) { /* file exists */
    fscanf(pidfile,"%d",(int *)&pid_old);
    if (kill(pid_old,0) != -1) {
      printf("File .PID exists, process running!\n");
      exit(0);
    } else {
      printf("File .PID exists, process not running!\n");
      fclose(pidfile);
      unlink(pidfilename);
    }
  }
  if ((pidfile = fopen(pidfilename, "w")) != NULL) {
    fprintf(pidfile, "%d\n", (int)pid);
    fclose(pidfile);
  } else {
    printf("Can't open file %s: %s\n", pidfilename, strerror(errno));
    exit(0);
  }

  told = tnow = time(NULL);
  while (1) {
    flock(utmp_file, LOCK_UN);
    sleep(scan_time);
    lseek(utmp_file, 0L, SEEK_SET);
    if (flock(utmp_file, LOCK_EX | LOCK_NB) < 0)
      if (errno == EWOULDBLOCK) {
	report(LOG_WARNING, "cannot lock utmp file");
	continue;
      }
    db = dbopen(dbfile, O_RDWR | O_EXLOCK, 0600, DB_HASH, NULL);
    if (!db) {
      report(LOG_ERR, "unable to open user database %s: %s", dbfile, strerror(errno));
      exit(EX_OSFILE);
    }
    tnow = time(NULL);
    while (read(utmp_file, &w, sizeof(w)) == sizeof(w)) {
      if (!*w.ut_user)
	continue;
      w.ut_user[UT_NAMESIZE-1] = '\0'; w.ut_line[UT_LINESIZE-1] = '\0'; w.ut_host[UT_HOSTSIZE-1] = '\0';
      if (db_update(db, told - tnow, &w) < 0)
	report(LOG_INFO, "no such user: %s", w.ut_user);
      if (reinit && (tnow - w.ut_time) > reinit) {
	report(LOG_INFO, "user %s on line %s nas %s logged in too long", w.ut_user, w.ut_line, w.ut_host);
	killuser(&w, 0);
      }
#ifdef STRICT_TIME_CHECK
      if (strict_time_check(w.ut_user, w.ut_host, w.ut_line)) {
	report(LOG_INFO, "user %s on line %s nas %s time interval is over", w.ut_user, w.ut_line, w.ut_host);
	killuser(&w, 0);
      }
#endif
    }
    told = tnow;
    (*db->close)(db);
    db = NULL;
  }
  exit(EX_OSERR);
}
